RC6_ROUNDS      = 20
RC6_SIZEOFS     = (RC6_ROUNDS*2+4)*4
RC6_MAGICP      = 0B7E15163h
RC6_MAGICQ      = 09E3779B9h

data section
#include "D:\Programming\ASM Compilers\RADasm\GoAsm\Inc\Windows.inc"
#define CALG_MD5 08003h
#define CALG_SHA 08004h
#define CRYPT_EXPORTABLE 1
#define CRYPT_VERIFYCONTEXT 0F0000000h
#define HP_HASHVAL 2  // Hash value
#define PROV_RSA_FULL 1

CODE SECTION

RC6EncryptFile FRAME pszFile, pszOutfile, pszKey
	uses edi,esi,ebx
	LOCAL hFile					:D
	LOCAL pFileData				:D
	LOCAL cbFile				:D
	LOCAL cbRead				:D
	LOCAL pHashKey				:D
	LOCAL SystemInfo			:SYSTEM_INFO
	LOCAL SArray[RC6_SIZEOFS]	:B

	invoke szLen,[pszKey]
	test eax,eax
	jz >>.ERROR_KEY_LEN

	// Use RC6 to encrypt a file
	// The actual key can be any length however it will be hashed to 16 bytes
	// using the SHA hash from the Crypt API this makes for better control of
	// the key length and still provides adequate security.
	// The 16 encrypted bytes from a block will be the key to the next block

	invoke GetSystemInfo,offset SystemInfo
	invoke VirtualAlloc,NULL,32,MEM_COMMIT,PAGE_READWRITE
	mov [pHashKey],eax

	invoke rc6_generatekey,[pszKey],[pHashKey]
	test eax,eax
	js >>.ERROR_KEY

	// We now have a key at pHashKey, we can begin to encrypt the file

	// Open the unencrypted file
	invoke CreateFile,[pszFile],GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0
	mov [hFile], eax
	test eax,eax
	js >>.ERROR_KEY
	
	invoke GetFileSize,[hFile],0
	mov [cbFile],eax
	test eax,eax
	jz >>.ERROR_SIZE

	// Round up to the next page using the system granularity
	// to calculate the next higher page boundary
	mov ecx,[SystemInfo.dwPageSize]
	dec ecx
	add eax,ecx
	not ecx
	and eax,ecx

	// Allocate a buffer for the file data
	invoke VirtualAlloc,NULL,eax,MEM_COMMIT,PAGE_READWRITE
	mov [pFileData],eax

	invoke ReadFile,[hFile],eax,[cbFile],offset cbRead,0

	// We're done reading so close the file
	invoke CloseHandle,[hFile]

	// Calculate the number of steps necessary to get through the entire file
	mov ebx,[cbFile]
	add ebx,0Fh
	and ebx,0FFFFFFF0h
	mov [cbFile],ebx
	shr ebx,4

	// Step through the file encrypting it
	mov esi,[pFileData]
	:
		invoke rc6_init,offset SArray,[pHashKey],16
		invoke rc6_encrypt,offset SArray,esi
		invoke memcopy,esi,[pHashKey],16
		add esi,16
		dec ebx
		jnz <

	invoke GetFileExists,[pszOutfile]
	mov ecx,TRUNCATE_EXISTING
	mov edx,CREATE_NEW
	test eax,eax
	cmovz ecx,edx

	// Save the buffer to the encrypted file
	invoke CreateFile,[pszOutfile],GENERIC_WRITE,FILE_SHARE_WRITE,0,ecx,0,0
	mov [hFile],eax
	invoke WriteFile,[hFile],[pFileData],[cbFile],offset cbRead,0
	invoke CloseHandle,[hFile]

	invoke VirtualFree,[pHashKey],NULL,MEM_RELEASE
	invoke VirtualFree,[pFileData],NULL,MEM_RELEASE
	RET

	.ERROR_SIZE
	invoke CloseHandle,[hFile]
	.ERROR_KEY
	invoke VirtualFree,[pHashKey],NULL,MEM_RELEASE
	.ERROR_KEY_LEN
	xor eax,eax
	dec eax ; Return error (-1)
	RET

ENDF

RC6DecryptFile FRAME pszFile, pszOutfile, pszKey
	uses edi,esi,ebx
	LOCAL hFile					:D
	LOCAL pFileData				:D
	LOCAL cbFile				:D
	LOCAL cbRead				:D
	LOCAL pHashKey				:D
	LOCAL SystemInfo			:SYSTEM_INFO
	LOCAL SArray[RC6_SIZEOFS]	:B

	invoke szLen,[pszKey]
	test eax,eax
	jz >>.ERROR_KEY_LEN

	// Use RC6 to decrypt a file
	// The actual key can be any length however it will be hashed to 16 bytes
	// using the SHA hash from the Crypt API this makes for better control of
	// the key length and still provides adequate security.
	// The 16 encrypted bytes from a block will be the key to the next block

	invoke GetSystemInfo,offset SystemInfo
	invoke VirtualAlloc,NULL,32,MEM_COMMIT,PAGE_READWRITE
	mov [pHashKey],eax

	invoke rc6_generatekey,[pszKey],[pHashKey]
	test eax,eax
	js >>.ERROR_KEY

	// We now have a key at pHashKey, we can begin to encrypt the file

	// Open the encrypted file
	invoke CreateFile,[pszFile],GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0
	mov [hFile], eax
	test eax,eax
	js >>.ERROR_KEY
	
	invoke GetFileSize,[hFile],0
	mov [cbFile],eax
	test eax,eax
	jz >>.ERROR_SIZE

	// Round up to the next page using the system granularity
	// to calculate the next higher page boundary
	mov ecx,[SystemInfo.dwPageSize]
	dec ecx
	add eax,ecx
	not ecx
	and eax,ecx

	// Allocate a buffer for the file data
	invoke VirtualAlloc,NULL,eax,MEM_COMMIT,PAGE_READWRITE
	mov [pFileData],eax

	invoke ReadFile,[hFile],eax,[cbFile],offset cbRead,0

	// We're done reading so close the file
	invoke CloseHandle,[hFile]

	// Calculate the number of steps necessary to get through the entire file
	mov ebx,[cbFile]
	add ebx,0Fh
	and ebx,0FFFFFFF0h
	mov [cbFile],ebx
	shr ebx,4

	// Step through the file decrypting it
	mov esi,[pFileData]
	:
		invoke rc6_init,offset SArray,[pHashKey],16
		invoke memcopy,esi,[pHashKey],16
		invoke rc6_decrypt,offset SArray,esi
		add esi,16
		dec ebx
		jnz <

	invoke GetFileExists,[pszOutfile]
	mov ecx,TRUNCATE_EXISTING
	mov edx,CREATE_NEW
	test eax,eax
	cmovz ecx,edx

	// Save the buffer to the encrypted file
	invoke CreateFile,[pszOutfile],GENERIC_WRITE,FILE_SHARE_WRITE,0,ecx,0,0
	mov [hFile],eax
	invoke WriteFile,[hFile],[pFileData],[cbFile],offset cbRead,0
	invoke CloseHandle,[hFile]

	invoke VirtualFree,[pHashKey],NULL,MEM_RELEASE
	invoke VirtualFree,[pFileData],NULL,MEM_RELEASE
	RET

	.ERROR_SIZE
	invoke CloseHandle,[hFile]
	.ERROR_KEY
	invoke VirtualFree,[pHashKey],NULL,MEM_RELEASE
	.ERROR_KEY_LEN
	xor eax,eax
	dec eax ; Return error (-1)
	RET

ENDF

rc6_generatekey FRAME pszKey,pHashKey
	uses edi,esi,ebx
	LOCAL dwStatus				:D
	LOCAL bResult				:D
	LOCAL hProv					:D
	LOCAL hHash					:D
	LOCAL cbRead				:D
	LOCAL rgbHash[64]			:B
	LOCAL cbHash				:D
	LOCAL rgbDigits[16]			:B

	// Generate a 16 byte SHA HASH from the key
	mov [rgbDigits],"0123"
	mov [rgbDigits+4],"4567"
	mov [rgbDigits+8],"89ab"
	mov [rgbDigits+12],"cdef"

	invoke szLen,[pszKey]
	test eax,eax
	jz >>.ERRORCONTEXT
	mov [cbRead],eax

	invoke CryptAcquireContext,offset hProv,NULL,NULL,PROV_RSA_FULL,CRYPT_VERIFYCONTEXT
	test eax,eax
	jz >>.ERRORCONTEXT

	invoke CryptCreateHash,[hProv], CALG_SHA, NULL, NULL, offset hHash
	test eax,eax
	jz >>.ERRORHASH

	invoke CryptHashData,[hHash], [pszKey], [cbRead], NULL
	test eax,eax
	jz >>.ERROROTHER

	// The hash is created zero the key so it doesn't sit in memory
	mov edi,[pszKey]
	mov ecx,[cbRead]
	xor eax,eax
	rep stosb

	mov D[cbHash],64
	invoke CryptGetHashParam,[hHash], HP_HASHVAL, offset rgbHash, offset cbHash, NULL
	test eax,eax
	jz >>.ERROROTHER

	// convert the hash to an SHA string using a lookup table
	xor eax,eax
	xor edx,edx
	mov ebx,[pHashKey]
	mov edi,offset rgbHash
	mov ecx,[cbHash]
	mov esi,offset rgbDigits
	:
		mov al,[edi]
		shr al,4
		mov dl,[esi+eax]
		mov [ebx],dl
		inc ebx
		mov al,[edi]
		and al,0fh
		mov dl,[esi+eax]
		mov [ebx],dl
		inc edi
		inc ebx
		dec ecx
		jnz <
	mov B[ebx],0

	invoke CryptDestroyHash [hHash]
	invoke CryptReleaseContext, [hProv], NULL
	xor eax,eax
	ret

	.ERROROTHER
	invoke CryptDestroyHash [hHash]

	.ERRORHASH
	invoke CryptReleaseContext, [hProv], NULL

	.ERRORCONTEXT
	invoke VirtualFree,[pHashKey],NULL,MEM_RELEASE
	xor eax,eax
	dec eax
	RET
ENDF

/*
RC6 implementation by TOLA, MASM translation by Webring, GoAsm translation by Donkey
*/

rc6_encrypt FRAME lpS, lpBuffer
	USES ebx, esi, edi
	LOCAL   Counter:D
	LOCAL	ESPSave:D

	mov     edi,[lpBuffer]
	mov     esi,[lpS]
	mov     [ESPSave],esp

	mov     eax,[edi]
	mov     ebx,[edi+4]
	mov     ecx,[edi+8]
	mov     edx,[edi+12]

	add     ebx,[esi]
	add     edx,[esi+4]

	xor     edi,edi
	inc     edi
	mov     [Counter],edi
.loop:

	lea     edi,[ebx*2+1]
	lea     esp,[edx*2+1]
	imul    edi,ebx
	imul    esp,edx
	rol     edi,5
	rol     esp,5

	xchg    ecx,esp
	xor     eax,edi
	rol     eax,cl
	mov     esi,[Counter]
	shl     esi,3
	add     esi,[lpS]
	xchg    ecx,esp
	add     eax,[esi]

	xchg    ecx,edi
	xor     edi,esp
	rol     edi,cl
	add     edi,[esi+4]
	xchg    ecx,edi

	xchg    eax,ebx
	xchg    ecx,edx
	xchg    ebx,edx

	inc     D[Counter]
	cmp     D[Counter],RC6_ROUNDS
	jbe     <.loop
	mov     esi,[lpS]
	mov     edi,[lpBuffer]

	add     eax,[esi+2*RC6_ROUNDS*4+2*4]

	add     ecx,[esi+2*RC6_ROUNDS*4+3*4]

	mov     [edi],eax
	mov     [edi+4],ebx
	mov     [edi+8],ecx
	mov     [edi+12],edx
	mov     esp,[ESPSave]
	ret
ENDF

rc6_decrypt FRAME lpS, lpBuffer
	USES ebx, esi, edi
	LOCAL   Counter:D
	LOCAL	ESPSave:D

	mov     edi,[lpBuffer]
	mov     esi,[lpS]
	mov     [ESPSave],esp

	mov     eax,[edi]
	mov     ebx,[edi+4]
	mov     ecx,[edi+8]
	mov     edx,[edi+12]

	sub     ecx,[esi+2*RC6_ROUNDS*4+3*4]
	sub     eax,[esi+2*RC6_ROUNDS*4+2*4]

	mov     edi,RC6_ROUNDS
	mov     [Counter],edi
.loop:

	xchg    eax,edx
	xchg    ecx,ebx
	xchg    edx,ebx

	lea     esp,[edx*2+1]
	imul    esp,edx
	rol     esp,5

	lea     edi,[ebx*2+1]
	imul    edi,ebx
	rol     edi,5

	xchg    ecx,edi
	mov     esi,[Counter]
	shl     esi,3
	add     esi,[lpS]
	sub     edi,[esi+4]
	ror     edi,cl
	xor     edi,esp
	xchg    ecx,edi

	sub     eax,[esi]
	xchg    ecx,esp
	ror     eax,cl
	xor     eax,edi
	xchg    ecx,esp

	dec     D[Counter]
	jnz     < .loop
	mov     esi,[lpS]
	mov     edi,[lpBuffer]

	sub     edx,[esi+4]

	sub     ebx,[esi]

	mov     [edi],eax
	mov     [edi+4],ebx
	mov     [edi+8],ecx
	mov     [edi+12],edx
	mov     esp,[ESPSave]
	ret
ENDF

rc6_init FRAME lpS, lpKey, dwKeysize
	USES esi, ebx, edi
	LOCAL   ESPSave:D
	LOCAL	Counter:D

	mov     eax,RC6_MAGICP
	mov     edi,[lpS]
	mov     [ESPSave],esp

	mov     [edi],eax

	mov     edx,RC6_MAGICQ
	mov     ecx,RC6_ROUNDS*2+3
	add     edi,4
.loop1:

	add     eax,edx
	mov     [edi],eax
	add     edi,4
	dec     ecx
	jnz     <.loop1

	mov     eax,[dwKeysize]
	mov     edx,RC6_ROUNDS*2+4
	shr     eax,2
	cmp     eax,edx
	jae     >.v_mul3
	lea     eax,[edx*2+edx]
	jmp     >.v_ok
.v_mul3:
	lea     eax,[eax*2+eax]
.v_ok:
	mov     edi,[lpS]
	mov     [Counter],eax
	mov     esi,[lpKey]

	xor     eax,eax
	xor     ebx,ebx
	xor     ecx,ecx
	xor     edx,edx

.loop2:

	add     eax,[edi+ecx*4]
	add     eax,ebx
	rol     eax,3
	mov     [edi+ecx*4],eax

	mov     esp,ecx
	lea     ecx,[eax+ebx]
	add     ebx,[esi+edx*4]
	add     ebx,eax
	rol     ebx,cl
	mov     ecx,esp
	mov     [esi+edx*4],ebx

	inc     ecx
	cmp     ecx,RC6_ROUNDS*2+4
	jne     >.i_reset_ok
	xor     ecx,ecx
.i_reset_ok:

	mov     esp,[dwKeysize]
	inc     edx
	shr     esp,2
	cmp     edx,esp
	jne     >.j_reset_ok
	xor     edx,edx
.j_reset_ok:
	dec     D[Counter]
	jnz     <.loop2
	mov     esp,[ESPSave]
	ret
ENDF
